/* process's Modules library
 * (c)1999 by Marek Zelem
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <mmm/modules.h>

int	modules_lib_version=0x0010;
char	modules_lib_strversion[]=
			"Modules Library ver.1.0 (c)17.8.1998 by Marek Zelem";

mod_syms_t modules_default_export_syms[]={
	SD(modules_lib_version),
	SD(modules_lib_strversion),
	SF(modules_inc_usecount),
	SF(modules_dec_usecount),
	SF(modules_add_syms),
	SF(modules_get_symbol),
	SF(modules_put_symbol),
	SF(modules_find_module),
	SE
					 };

int modules_verlib( char *buf )
{
	if( buf!=NULL )
		strcpy(buf,modules_lib_strversion);
	return(modules_lib_version);
}

static void del_use( struct s_moduse *mu );

static void free_modules_t( modules_t *m )
{ int i;
  struct s_moduse *mu;
	while( (mu=m->use)!=NULL )
		del_use(mu);
	while( (mu=m->used)!=NULL )
		del_use(mu);
	if( m->name!=NULL )	free(m->name);
	if( m->symnames!=NULL )
		elf_unmap(&(m->map_symnames));
	if( m->sym!=NULL )	free(m->sym);
	if( m->segment!=NULL )
	{	for(i=0;i<m->nr_seg;i++)
		{	if(m->segment[i].name!=NULL)
				free(m->segment[i].name);
			if(m->segment[i].base!=NULL)
				elf_unmap(&(m->segment[i].map));
		}
		free(m->segment);
	}
	if( m->undef!=NULL )	free(m->undef);
	if( m->rel!=NULL )	free(m->rel);
	free(m);
}

modules_t *modules_start( mod_syms_t *syms )
{ modules_t *mod;
  u_long i,n;
	if( (mod=malloc(sizeof(modules_t)))==NULL )
		return(NULL);
	mod->next=NULL;
	mod->prev=NULL;
	mod->use=NULL;
	mod->used=NULL;
	mod->usecount=0;
	mod->status=MS_INITIALIZED;
	mod->nr_sym=0;
	mod->nr_seg=0;
	mod->map_symnames.map_addr=NULL;
	mod->name=NULL; mod->symnames=NULL; mod->sym=NULL; mod->segment=NULL;
	mod->undef=NULL; mod->undef_len=0; mod->undef_size=0;
	mod->rel=NULL; mod->rel_len=0; mod->rel_size=0;
	n=0;
	while( modules_default_export_syms[n].name!=NULL )	n++;
	if( (mod->sym=malloc(n*sizeof(struct s_modsym)))==NULL )
		return(mod);
	mod->sym_size=n;
	for(i=0;i<n;i++)
	{	mod->sym[i].name=modules_default_export_syms[i].name;
		mod->sym[i].value=(long)(modules_default_export_syms[i].addr);
		mod->sym[i].size=0;
		mod->sym[i].segment=(u_short)-1;
		mod->sym[i].type=0;
	}
	mod->nr_sym=n;
	if( syms!=NULL )
		modules_add_syms(mod,syms);
	return(mod);
}

static int mod_remove_module( modules_t *mod, modules_t *m );

int modules_stop( modules_t *mod )
{ modules_t *m;
	if( mod->prev!=NULL )
		return(E_NOMOD);
	m=mod;
	while( (m=m->next)!=NULL )
	{	if( mod_remove_module(mod,m)>=0 )
			m=mod;
	}
	if( mod->next!=NULL )
	{	printf("modules_stop: Can't remove modules:");
		while( (m=mod->next)!=NULL )
		{	if( (m->prev->next=m->next)!=NULL )
				m->next->prev=m->prev;
			printf(" %s",m->name);
			free_modules_t(m);
		}
		printf("\n");
	}
	free_modules_t(mod);
	return(0);
}

static void link_new_symbols( modules_t *mod, modules_t *new );

int modules_add_syms( modules_t *mod, mod_syms_t *syms )
{ u_long i,n,p;
  void *new;
  modules_t *head;
	if( syms!=NULL )
	{	n=0;
		while( syms[n].name!=NULL )	n++;
		if( ( new=realloc(mod->sym,(n+mod->nr_sym)
				*sizeof(struct s_modsym)) )==NULL )
			return(-1);
		mod->sym=new;
		mod->sym_size=n+mod->nr_sym;
		p=mod->nr_sym;
		for(i=0;i<n;i++)
		{	mod->sym[p+i].name=syms[i].name;
			mod->sym[p+i].value=(long)(syms[i].addr);
			mod->sym[p+i].size=0;
			mod->sym[p+i].segment=(u_short)-1;
			mod->sym[p+i].type=0;
		}
		mod->nr_sym+=n;
		head=mod;
		while( head->prev!=NULL )	head=head->prev;
		link_new_symbols(head,mod);
	}
	return(0);
}

static int load_segments( modules_t *m, elf_t *elf )
{ int s,n;
  char *sstr=NULL;
	n=0;
	elf_for_each_section(elf,s)
	{	if( (elf_sect(elf,s).sh_flags & SHF_ALLOC)
//			&& elf_sect(elf,s).sh_type!=SHT_NOBITS 
			&& elf_sect(elf,s).sh_size>0
			&& elf_sect(elf,s).sh_addr==0 )
				n++;
	}
	if( n==0 )	return(0);
	if( (m->segment=malloc(n*sizeof(struct s_modseg)))==NULL )
		return(-1);
	m->nr_seg=n;
	sstr=elf_read_section(elf,elf->head.e_shstrndx);
	n=0;
	elf_for_each_section(elf,s)
	{	if( (elf_sect(elf,s).sh_flags & SHF_ALLOC)
//			&& elf_sect(elf,s).sh_type!=SHT_NOBITS
			&& elf_sect(elf,s).sh_size>0
			&& elf_sect(elf,s).sh_addr==0 )
		{	m->segment[n].map.map_addr=NULL;
			m->segment[n].name=strdup(
			 (sstr==NULL ? "???" : sstr+elf_sect(elf,s).sh_name) );
			m->segment[n].prot=PROT_READ;
			if( elf_sect(elf,s).sh_flags & SHF_EXECINSTR )
				m->segment[n].prot|=PROT_EXEC;
			if( elf_sect(elf,s).sh_flags & SHF_WRITE )
				m->segment[n].prot|=PROT_WRITE;
			m->segment[n].section=s;
			m->segment[n].base=
				elf_map_section(elf,s,&(m->segment[n].map),
					(m->segment[n].prot)|PROT_WRITE );
			if( m->segment[n].base==NULL )
			{	free(m->segment[n].name);
				if( sstr!=NULL )	free(sstr);
				m->nr_seg=n;
printf("load_segments: cant elf_map_sect: sect=%d\n",s);
				return(-1);
			}
			m->segment[n].size=m->segment[n].map.len;
			n++;
		}
	}
	if( sstr!=NULL )	free(sstr);
	return(0);
}

static int inc_use( modules_t *use, modules_t *used )
{ struct s_moduse *mu;
	if( use==used )	return(0);
	mu=use->use;
	while( mu!=NULL )
	{	if( mu->module_used==used )
		{	mu->count++;
			return(0);
		}
		mu=mu->next_use;
	}
	if( (mu=malloc(sizeof(struct s_moduse)))==NULL )
		return(-1);
	mu->count=1;
	mu->module_use=use;
	mu->module_used=used;
	mu->prev_used=NULL;
	if( (mu->next_used=used->used)!=NULL )
		mu->next_used->prev_used=mu;
	used->used=mu;
	mu->prev_use=NULL;
	if( (mu->next_use=use->use)!=NULL )
		mu->next_use->prev_use=mu;
	use->use=mu;
	return(0);
}

static void del_use( struct s_moduse *mu )
{	if( mu->prev_used!=NULL )
		mu->prev_used->next_used=mu->next_used;
	else
		mu->module_used->used=mu->next_used;
	if( mu->next_used!=NULL )
		mu->next_used->prev_used=mu->prev_used;
	if( mu->prev_use!=NULL )
		mu->prev_use->next_use=mu->next_use;
	else
		mu->module_use->use=mu->next_use;
	if( mu->next_use!=NULL )
		mu->next_use->prev_use=mu->prev_use;
	free(mu);
}

static int dec_use( modules_t *use, modules_t *used )
{ struct s_moduse *mu;
	if( use==used )	return(0);
	mu=use->use;
	while( mu!=NULL )
	{	if( mu->module_used==used )
		{	mu->count--;
			if( mu->count==0 )
				del_use(mu);
			return(0);
		}
		mu=mu->next_use;
	}
	return(-1);
}

struct s_modsym *modules_get_symbol( modules_t *me, modules_t *mod,
					char *name )
{ modules_t *m;
  int i;
	if( name==NULL || name[0]==0 )
		return(NULL);
	m=mod->next;
	if( mod->prev!=NULL )	m=NULL;
	while( 1 )
	{	if( m==NULL )	m=mod;
		if( m->sym!=NULL && (m==mod || m->symnames!=NULL) )
		{	for(i=0;i<m->nr_sym;i++)
			{	if( !strcmp(name,m->sym[i].name) )
				{	if( inc_use(me,m)<0 )
						return(NULL);
					return( &(m->sym[i]) );
				}
			}
		}
		if( m==mod )	break;
		else		m=m->next;
	}
	return(NULL);
}

int modules_put_symbol( modules_t *me, modules_t *mod, struct s_modsym *sym )
{ modules_t *m;
  int i;
	m=mod->next;
	if( mod->prev!=NULL )	m=NULL;
	while( 1 )
	{	if( m==NULL )	m=mod;
		if( m->sym!=NULL && (m==mod || m->symnames!=NULL) )
		{	i=(sym)-(m->sym);
			if( i>=0 && i<m->nr_sym )
			{	dec_use(me,m);
				return(0);
			}
		}
		if( m==mod )	break;
		else		m=m->next;
	}
printf("modules_put_symbol: symbol not exist!!!\n");
	return(-1);
}

static int link_symbols( modules_t *m, elf_t *elf, modules_t *mod,
		Elf32_Sym *sym, int nr_syms, int symtab )
{ int s,a,si;
  struct s_modseg *seg;
  struct s_modsym *msym;
  	sym[0].st_value=0;	/* first symbol is not valid !!! */
	for(s=1;s<nr_syms;s++)
	{	a=sym[s].st_shndx;
		if( a>0 && a<SHN_LORESERVE )
		{	if( (seg=m->segment)==NULL )
				goto Undef;
			for(si=0;si<m->nr_seg && seg[si].section!=a;)	si++;
			seg+=si;
			if(si>=m->nr_seg || seg->section!=a || seg->base==NULL)
				goto Undef;
			else
			{	sym[s].st_value+=(long)(seg->base);
				if( ELF32_ST_BIND(sym[s].st_info)==STB_GLOBAL )
				{  if( m->sym==NULL )
				   {	m->sym_size=16;
					m->nr_sym=0;
					if( (m->sym=malloc(m->sym_size
					   *sizeof(struct s_modsym)))==NULL )
						return(E_OTHER);
				   }
				   if( m->nr_sym>=m->sym_size )
				   {	m->sym_size+=16;
					if( (m->sym=realloc(m->sym,m->sym_size
					   *sizeof(struct s_modsym)))==NULL )
						return(E_OTHER);
				   }
				   m->sym[m->nr_sym].name=(m->symnames)+
				   			(sym[s].st_name);
				   m->sym[m->nr_sym].value=sym[s].st_value;
				   m->sym[m->nr_sym].size=sym[s].st_size;
				   m->sym[m->nr_sym].segment=si;
			 	   m->sym[m->nr_sym].type=0;
				   m->nr_sym++;
				}
			}
		}
		else if( a==0 )
		{	if( !strcmp("__modules_head",
				(m->symnames)+(sym[s].st_name)) )
				sym[s].st_value=(long)mod;
			else if( !strcmp("__this_module",
				(m->symnames)+(sym[s].st_name)) )
				sym[s].st_value=(long)m;
			else if( (msym=modules_get_symbol( m, mod,
				(m->symnames)+(sym[s].st_name) ))==NULL )
				goto Undef;
			else
				sym[s].st_value=msym->value;
		}
		else
Undef:		{	sym[s].st_value=0;      /* undefined */
			if( a==0 )
			{  if( m->undef==NULL )
			   {	m->undef_size=4;
				m->undef_len=0;
				if( (m->undef=malloc(m->undef_size
					   *sizeof(struct s_modsym)))==NULL )
					return(E_OTHER);
			   }
			   if( m->undef_len>=m->undef_size )
			   {	m->undef_size+=4;
				if( (m->undef=realloc(m->undef,m->undef_size
					   *sizeof(struct s_modsym)))==NULL )
					return(E_OTHER);
			   }
			   m->undef[m->undef_len].name=(m->symnames)+
				   			(sym[s].st_name);
			   m->undef[m->undef_len].value=0;
			   m->undef[m->undef_len].size=0;
			   m->undef[m->undef_len].segment=0;
			   m->undef[m->undef_len].type=0;
			   m->undef_len++;
			}
		}
	}
	return(0);
}

static int reloc_module( modules_t *m, elf_t *elf, modules_t *mod,
		Elf32_Sym *sym, int nr_syms, int symtab )
{ int s,a,i;
  struct s_modseg *seg;
  Elf32_Rel *rel=NULL;
  Elf32_Rela *rela=NULL;
  elfmap_t map;
  int rels=0;
  long *pos;
  int sy,rtype;
  long addend;
  	map.map_addr=NULL;
	elf_for_each_section(elf,s)
	{	if( elf_sect(elf,s).sh_type==SHT_REL
			|| elf_sect(elf,s).sh_type==SHT_RELA )
		{	if( elf_sect(elf,s).sh_link!=symtab )
			{
				printf("Bad symbol table!\n");
				return(E_BADVERSION);
			}
			a=elf_sect(elf,s).sh_info;
			if( (seg=m->segment)==NULL )
				return(E_OTHER);
			for(i=0;i<m->nr_seg && seg[i].section!=a;)	i++;
			seg+=i;
			if( i>=m->nr_seg || seg->section!=a )
			{
//				printf("Segment not allocated!\n");
				continue;
			}
			if( elf_sect(elf,s).sh_type==SHT_REL )
			{  if( elf_sect(elf,s).sh_entsize!=sizeof(Elf32_Rel) )
Bad_entsize:		   {	printf("Bad rel(a) entsize!\n");
				continue;
			   }
			   if( (rel=elf_map_section(elf,s,&map,PROT_READ))
					==NULL )
				return(E_OTHER);
			   rels=map.len/sizeof(Elf32_Rel);
			}
			else
			{  if( elf_sect(elf,s).sh_entsize!=sizeof(Elf32_Rela) )
				goto Bad_entsize;
			   if( (rela=elf_map_section(elf,s,&map,PROT_READ))
					   ==NULL )
				   return(E_OTHER);
			   rels=map.len/sizeof(Elf32_Rela);
			}
			for(i=0;i<rels;i++)
			{	if( rel!=NULL )
				{	pos=(long*)
					    ((seg->base)+(rel[i].r_offset));
					sy=ELF32_R_SYM(rel[i].r_info);
					rtype=ELF32_R_TYPE(rel[i].r_info);
					addend=*pos;
				}
				else
				{	pos=(long*)
					    ((seg->base)+(rela[i].r_offset));
					sy=ELF32_R_SYM(rela[i].r_info);
					rtype=ELF32_R_TYPE(rela[i].r_info);
					addend=rela[i].r_addend;
				}
				if( sy>=nr_syms || sym[sy].st_value==0 )
				{ int us;
					if( m->rel==NULL )
					{ m->rel_size=4;
					  m->rel_len=0;
					  m->rel=malloc(m->rel_size
						*sizeof(struct s_modrel));
					}
					if( m->rel_len>=m->rel_size )
					{ m->rel_size=16+m->rel_len;
					  m->rel=realloc(m->rel,m->rel_size
						*sizeof(struct s_modrel));
					}
					us=0;
					if( m->undef==NULL )
						goto Undef;
					while( us<m->undef_len )
					{ if( m->rel!=NULL
					    && !strcmp(m->undef[us].name,
					    (m->symnames)+(sym[sy].st_name))  )
					  { m->rel[m->rel_len].addr=pos;
					    m->rel[m->rel_len].addend=addend;
					    m->rel[m->rel_len].symbol=us;
					    m->rel[m->rel_len].type=rtype;
					    *pos=0;
					    m->rel_len++;
					    goto Next;
					  }
					  us++;
					}
Undef:
printf("undefined symbol(%d): \"%s\"\n",sy,(m->symnames)+(sym[sy].st_name));
					goto Err;
Next:					continue;
				}
				switch( rtype )
				{ case R_386_32:
					*pos=(sym[sy].st_value) + addend;
					break;
				  case R_386_PC32:
					*pos=(sym[sy].st_value) + addend - 
						(long)pos;
					break;
				  case R_386_GLOB_DAT:
				  case R_386_JMP_SLOT:
					*pos=(sym[sy].st_value);
					break;
				  case R_386_RELATIVE:
					*pos=((long)(seg->base)) + addend;
					break;
				  default:
					printf("UnImplemented reloc type!\n");
Err:					if( rel!=NULL )
					{  elf_unmap(&map);  rel=NULL;  }
					if( rela!=NULL )
					{  elf_unmap(&map); rela=NULL; }
					return(E_UNDEF);
				}
//printf("rel: %s = %lx\n",(m->symnames)+(sym[sy].st_name),*pos);
			}
			if( rel!=NULL )
			{	elf_unmap(&map);	rel=NULL;	}
			if( rela!=NULL )
			{	elf_unmap(&map);	rela=NULL;	}
		}
	}
	return(0);
}

static int mod_init_module( modules_t *mod, modules_t *m );

static void link_new_symbols( modules_t *mod, modules_t *new )
{ modules_t *m;
  struct s_modsym *sym;
  int i,changed;
  long *pos;
	for(m=mod->next;m!=NULL;m=m->next)
	{	if( m==new )		continue;
		if( m->undef==NULL )	continue;
		if( m->rel==NULL )	continue;
		changed=0;
		for(i=0;i<m->undef_len;i++)
		{	if( (sym=modules_get_symbol(m,new,m->undef[i].name))
				   !=NULL )
			{	m->undef[i].value=sym->value;
				changed=1;
			}
		}
		if( !changed )		continue;
		for(i=0;i<m->rel_len;i++)
		{	if( m->rel[i].type & 0x4000 )	continue;
			sym=&(m->undef[m->rel[i].symbol]);
			if( sym->value!=0 )
			{	pos=m->rel[i].addr;
				switch( m->rel[i].type )
				{ case R_386_32:
					*pos=(sym->value) + m->rel[i].addend;
					break;
				  case R_386_PC32:
					*pos=(sym->value) + m->rel[i].addend -
						(long)pos;
					break;
				  case R_386_GLOB_DAT:
				  case R_386_JMP_SLOT:
					*pos=(sym->value);
					break;
/*  nemam base	!!!		  case R_386_RELATIVE:
    zistit prehladanim segmentov	*pos=((long)(seg->base)) + addend;
					break;
*/
				  default:
					printf("UnImplemented reloc type!\n");
					continue;
				}
				m->rel[i].type|=0x4000;
			}
		}
		if( (m->status & MS_AUTOINIT) && !(m->status & MS_INITIALIZED) )
			mod_init_module(mod,m);
	}
}

//static void segments_hexdump( modules_t *m );
static int call_module( modules_t *mod, modules_t *m, char *func, int r );
	
int modules_load_module( modules_t *mod, const char *filename,
	       	const char *name, int status )
{ elf_t *elf=NULL;
  int r,symtab;
  modules_t *m=NULL;
  Elf32_Sym *sym=NULL;
  elfmap_t sym_map;
  	if( name==NULL || name[0]==0 )
		name=filename;
	if( strrchr(name,'/')!=NULL )
		name=strrchr(name,'/')+1;
	if( modules_find_module(mod,name)!=NULL )
		return(E_MODEXIST);
  	sym_map.map_addr=NULL;
	if( (elf=elf_open(filename))==NULL )
		return(elf_errno);
	if( (r=elf_test(elf))<0 )	goto Err;
	if( elf->head.e_type!=ET_REL )
	{	r=E_BADTYPE;	goto Err;	}
	if( (m=malloc(sizeof(modules_t)))==NULL )
	{	r=E_OTHER;	goto Err;	}
	m->use=NULL;
	m->used=NULL;
	m->usecount=0;
	m->status= status & MS_SET_MASK;
	m->next=NULL;	m->prev=NULL;
	m->map_symnames.map_addr=NULL;
	m->name=NULL; m->symnames=NULL; m->sym=NULL; m->segment=NULL;
	m->undef=NULL; m->undef_len=0; m->undef_size=0;
	m->rel=NULL; m->rel_len=0; m->rel_size=0;
	if( (m->name=strdup(name))==NULL )
	{	r=E_OTHER;	goto Err;	}
	elf_for_each_section(elf,symtab)
	{	if( elf_sect(elf,symtab).sh_type==SHT_SYMTAB )
			break;
	}
	if( (sym=elf_map_section(elf,symtab,&sym_map,PROT_READ|PROT_WRITE))
			==NULL )
	{	r=E_BADTYPE;	goto Err;	}
	if( (m->symnames=elf_map_section(elf,elf_sect(elf,symtab).sh_link,
			              &(m->map_symnames),PROT_READ))==NULL )
	{	r=E_BADTYPE;	goto Err;	}
	m->sym=NULL;
	m->nr_sym=0;
	m->sym_size=0;
	m->segment=NULL;
	m->nr_seg=0;
	if( load_segments(m,elf)<0 )
	{	r=E_OTHER;	goto Err;	}
	if( (r=link_symbols(m,elf,mod,sym,sym_map.len/sizeof(Elf32_Sym),symtab))<0 )
		goto Err;
	if( (r=reloc_module(m,elf,mod,sym,sym_map.len/sizeof(Elf32_Sym),symtab))<0 )
		goto Err;
//segments_hexdump(m);
	r=0;
	if( m->rel!=NULL )
		r=E_UNDEF;
	r=call_module(mod,m,"__loaded_module",r);
	if( r<0 )	goto Err;
	if( (m->next=mod->next)!=NULL )
		m->next->prev=m;
	m->prev=mod;
	mod->next=m;
	r=0;
	link_new_symbols(mod,m);
	if( (m->status & MS_AUTOINIT) && !(m->status & MS_INITIALIZED) )
		mod_init_module(mod,m);
	goto Koniec;

Err:	if( r==E_UNDEF )
		modules_print_undef(0,m);
	if(m!=NULL)
		free_modules_t(m);
Koniec:	if( sym!=NULL )		elf_unmap(&(sym_map));
	if( elf!=NULL )		elf_close(elf);
	return(r);
}

modules_t *modules_find_module( modules_t *mod, const char *name )
{
	if( strrchr(name,'/')!=NULL )
		name=strrchr(name,'/')+1;
	while( mod->prev!=NULL )	mod=mod->prev;
	mod=mod->next;
	while( mod!=NULL )
	{	if( !strcmp(mod->name,name) )
			return(mod);
		mod=mod->next;
	}
	return(NULL);
}

static int call_module( modules_t *mod, modules_t *m, char *func, int r )
{ struct s_modsym *msym;
	msym=modules_get_symbol(mod,m,func);
	if( msym!=NULL && msym->value!=0
		&& (m->segment[msym->segment].prot & PROT_EXEC) )
	{ int (*fn)(void);
		fn=(int(*)(void))msym->value;
		r=fn();
	}
	if( msym!=NULL )
		modules_put_symbol(mod,m,msym);
	return(r);
}

static int mod_init_module( modules_t *mod, modules_t *m )
{ int r,i;
	if( m->status & MS_INITIALIZED )	return(0);
	if( m->rel!=NULL )
	{	r=0;
		for(i=0;i<m->rel_len;i++)
		{	if( !(m->rel[i].type & 0x4000) )
			{	r=E_UNDEF;
				break;
			}
		}
		if( r<0 )
		{	r=call_module(mod,m,"__test_module",r);
			if( r<0 )	return(E_UNDEF);
		}
	}
	r=E_NOINIT;
	r=call_module(mod,m,"init_module",r);
	if( r>=0 )	m->status|=MS_INITIALIZED;
	return(r);
}

int modules_init_module( modules_t *mod, const char *name )
{ modules_t *m;
	if( (m=modules_find_module(mod,name))==NULL )
		return(E_NOMOD);
	return(mod_init_module(mod,m));
}

static int mod_remove_module( modules_t *mod, modules_t *m )
{ int r;
	if( m==NULL || m->prev==NULL )	return(E_IMPOSIBLE);
	if( m->used!=NULL )
		return(E_MODSYMUSED);
	if( m->usecount!=0 )
	{	r=E_MODINUSE;
		r=call_module(mod,m,"__finish_module",r);
		if( r<0 )	return(r);
	}
	if( m->used!=NULL )
		return(E_MODSYMUSED);
	if( m->usecount!=0 )
		return(E_MODINUSE);
	/* najprv ho odpojit */
	if( (m->prev->next=m->next)!=NULL )
		m->next->prev=m->prev;
	m->next=NULL;	m->prev=NULL;
	r=0;
	if( m->status & MS_INITIALIZED )
		r=call_module(mod,m,"cleanup_module",r);
	free_modules_t(m);
	return(0);
}

int modules_remove_module( modules_t *mod, const char *name )
{ modules_t *m;
	if( (m=modules_find_module(mod,name))==NULL )
		return(E_NOMOD);
	return(mod_remove_module(mod,m));
}

void modules_inc_usecount( modules_t *mod )
{
	mod->usecount++;
}

void modules_dec_usecount( modules_t *mod )
{
	if( mod->usecount>0 )
		mod->usecount--;
}


/* vypisy */

static void m_print_syms( int fd, struct s_modsym *s, long n )
{ long i;
  char buf[128];
	for(i=0;i<n;i++)
	{	snprintf(buf,128,"\t%s\t%lx\n",
			s[i].name,s[i].value);
		buf[127]=0;
		write(fd,buf,strlen(buf));
	}
}

int modules_print_undef( int fd, modules_t *mod )
{ long i;
  char buf[128];
	snprintf(buf,128,"undefined symbols:");
	write(fd,buf,strlen(buf));
	if( mod->undef!=NULL )
	{	for(i=0;i<mod->undef_len;i++)
		{	snprintf(buf,128," %s",mod->undef[i].name);
			write(fd,buf,strlen(buf));
		}
	}
	write(fd,"\n",1);
	return(0);
}

int modules_print( int fd, modules_t *mod, int level )
{ struct s_moduse *used;
  char buf[128];
  	if( fd==0 )	fd=1;
	while( mod!=NULL )
	{	snprintf(buf,128,"MODULE %s usecount=%d",
			mod->name,mod->usecount);
		write(fd,buf,strlen(buf));
		if( !(mod->status & MS_INITIALIZED) )
			write(fd," (uninitialized)",16);
		if( mod->status & MS_AUTOINIT )
			write(fd," (autoinit)",11);
		if( mod->status & MS_AUTOCLEAN )
			write(fd," (autoclean)",12);
		if( level>=1 )
		{  write(fd,"\nused by:",9);
		   used=mod->used;
		   while( used!=NULL )
		   {	snprintf(buf,128," %s(%d)",used->module_use->name,
					 used->count);
			write(fd,buf,strlen(buf));
			used=used->next_used;
		   }
		}
		if( level>=2 )
		{  write(fd,"\nuse:",5);
		   used=mod->use;
		   while( used!=NULL )
		   {	snprintf(buf,128," %s(%d)",used->module_used->name,
					 used->count);
			write(fd,buf,strlen(buf));
			used=used->next_use;
		   }
		}
		write(fd,"\n",1);
		if( level>=3 && mod->sym!=NULL )
		{	write(fd,"  Symbols:\n",11);
			m_print_syms(fd,mod->sym,mod->nr_sym);
		}
		if( level>=4 && mod->undef!=NULL )
		{	write(fd,"  Undefined symbols:\n",21);
			m_print_syms(fd,mod->undef,mod->undef_len);
		}
		mod=mod->next;
	}
	return(0);
}
/*
static void segments_hexdump( modules_t *m )
{ int s;
  u_char *base;
  long len,i,j;
	for(s=0;m->segment!=NULL && s<m->nr_seg;s++)
	{	base=(u_char*)(m->segment[s].base);
		len=m->segment[s].size;
		printf("Segment %s base=%p size=0x%lx\n",
			m->segment[s].name,base,len);
		for(j=0,i=0;i<len;j++,i++)
		{	if( j>=8 )
			{	printf("\n");	j=0;	}
			if( j==0 )	printf("0x%08lx ",i);
			printf(" 0x%02x",base[i]);
		}
		printf("\n");
	}
}
*/
