/* io_lib.c	  */
/* (c) by Marek Zelem */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include "io_lib.h"
#ifndef linux
#include <string.h>
#endif

int io_verlib( char *buf )
{
	if( buf!=NULL )
		strcpy(buf,"IO Library ver.2.0 (c)14.3.1998 by Marek Zelem");
	return(0x0020);
}

/* io */

#define timeval2msec(tv)	( ((tv).tv_usec)/1000 + ((tv).tv_sec)*1000 )

typedef struct	{
			unsigned short	generation;
			unsigned short	next_free;
			io_t	*io;
		} iotab_t;

typedef struct	{
			fd_set	*fds;
			iotab_t	*tab;
			int	tablen;
			unsigned short first_free;
			unsigned short last_free;
			long	last_evid;
		} iot_t;
#define	iott(io)	((iot_t*)((io)->ptr))

int _io_table_realoc( io_t *io, unsigned int tablen )
{ unsigned short i,f;
  iotab_t *old;
	if( tablen > 0x00ffff )	return(-1);
	old=iott(io)->tab;
	if( (iott(io)->tab=realloc(old,sizeof(iotab_t)*tablen))==NULL )
	{	iott(io)->tab=old;
		return(-1);
	}
	f=iott(io)->last_free;
	for(i=iott(io)->tablen;i<tablen;i++)
	{	iott(io)->tab[i].generation=0;
		iott(io)->tab[i].next_free=0xffff;
		if( f<tablen )	iott(io)->tab[f].next_free=i;
		else	iott(io)->first_free=i;
		f=i;
	}
	iott(io)->last_free=f;
	iott(io)->tablen=(unsigned short)tablen;
	return(0);
}

ioid_t _io_table_getfree( io_t *io, io_t *p )
{ unsigned short i;
  ioid_t id;
	io=io->head;
	if( (i=iott(io)->first_free) >= iott(io)->tablen )
		if( _io_table_realoc(io,(iott(io)->tablen)+128)<0 )
			return(0xffffffff);
	if( (i=iott(io)->first_free) < iott(io)->tablen )
	{	if( (iott(io)->first_free=iott(io)->tab[i].next_free)
		    >= iott(io)->tablen )
			iott(io)->last_free=0xffff;
		iott(io)->tab[i].io=p;
		iott(io)->tab[i].next_free=0xffff;
		id=iott(io)->tab[i].generation;
		id=(id<<16)|i;
		p->id=id;
		return( id );
	}
	return(0xffffffff);
}

int _io_table_setfree( io_t *io, ioid_t id )
{ unsigned short i,g,j;
	io=io->head;
	i= (unsigned short)(id&0x00ffff);
	g= (unsigned short)((id>>16)&0x00ffff);
	if( i >= iott(io)->tablen || iott(io)->tab[i].generation!=g )
		return(-1);
	iott(io)->tab[i].io=NULL;
	iott(io)->tab[i].generation++;
	iott(io)->tab[i].next_free=0xffff;
	if( (j=iott(io)->last_free) < iott(io)->tablen )
		iott(io)->tab[j].next_free=i;
	else
		iott(io)->first_free=i;
	iott(io)->last_free=i;
	return(0);
}

void _delete_evlist( io_t *io )
{ evlist_t *e,*ne;
	e=io->evlist;
	while( e!=NULL )
	{	ne=e->next;
		if( e->free!=NULL )
			e->free(e->arg);
		free(e);
		e=ne;
	}
	io->evlist=NULL;
}

io_t *io_init( int(*func)(io_t*,int,void*) )
{ io_t *io;
	if( (io=malloc(sizeof(io_t)+sizeof(iot_t)+sizeof(fd_set)))==NULL )
		return(NULL);
	io->head=io;	io->id=0xffffffff;
	io->next=io->prev=NULL;
	io->func=func; io->ufn=NULL; io->ptr=&io[1];
	io->evlist=NULL;
	iott(io)->fds=(fd_set*)(&(iott(io)[1]));
	iott(io)->tablen=0;
	iott(io)->last_free=0xffff;
	iott(io)->first_free=0xffff;
	iott(io)->last_evid=16;		/* prvych 16 rezervovanych */
	if( _io_table_realoc(io,128)<0 )
	{	free(io);
		return(NULL);
	}
	io->rd=io->wr=io->ex= -1;
	io->msec=io->nmsec=0;
	io->prior= -1;
	io->user1=io->user2=0;
	io->inbuf=io->outbuf=io->ptr2=NULL;
	io->caller=io->user_id=0xffffffff;
	FD_ZERO(iott(io)->fds);
	if( io->func!=NULL )
		if( io->func(io,IO_INIT,NULL)<0 )
		{	io->caller=0xffffffff;
			io->func(io,IO_DEL,NULL);
			free(iott(io)->tab);
			free(io);
			return(NULL);
		}
	return(io);
}

int io_final( io_t *io )
{ io_t *p,*q;
  int i;
	if( io==NULL || io->prev!=NULL )	return(-1);
	p=io;
	while( p->next!=NULL )	p=p->next;
	while( (q=p->prev)!=NULL )
	{	p->caller=0xffffffff;
		if( p->func!=NULL )	p->func(p,IO_DEL,NULL);
		_delete_evlist(p);
		_io_table_setfree(io,p->id);
		free(p);
		p=q;
	}
	if( io->func!=NULL )	io->func(io,IO_DEL,NULL);
	_delete_evlist(io);
	for(i=FD_SETSIZE-1;i>=0;i--)
		if( FD_ISSET(i,iott(io)->fds) )	close(i);	
	free(iott(io)->tab);
	free(io);
	return(0);
}

int io_closeall( io_t *io )
{ int i;
	if( io==NULL || io->prev!=NULL )	return(-1);
	for(i=FD_SETSIZE-1;i>=0;i--)
	{	if( FD_ISSET(i,iott(io)->fds) )
		{	shutdown(i,2);
			close(i);
		}
	}
	return(0);
}

io_t *io_first( io_t *io )
{
	return(io->head);
}

io_t *io_find( io_t *io, long id )
{ unsigned short i,g;
	io=io->head;
	i= (unsigned short)(id&0x00ffff);
	g= (unsigned short)((id>>16)&0x00ffff);
	if( i >= iott(io)->tablen || iott(io)->tab[i].generation!=g )
		return(NULL);
	return(iott(io)->tab[i].io);
}

ioid_t io_id( io_t *io )
{
	return( io->id );
}

int io_setopen(io_t *io, int fd)
{
	if( fd>=0 )
	{	io=io->head;
		FD_SET(fd,iott(io)->fds);
	}
	return(fd);
}

int io_setclose(io_t *io, int fd)
{
	if( fd>=0 )
	{	io=io->head;
		FD_CLR(fd,iott(io)->fds);
	}
	return(0);
}

int io_close(io_t *io, int fd)
{	io_setclose(io,fd);
	return(close(fd));
}

int _io_add(io_t *io, io_t *new, int prior)
{
	new->prior=prior;
	new->head=io->head;
	while( 1 )
	{
		if( io->prior < prior || io->prev==NULL )
		{	if( (io->next==NULL || io->next->prior>=prior) )
			{	if( (new->next=io->next)!=NULL )
					new->next->prev=new;
				new->prev=io;
				io->next=new;
				return(1);
			}
			else
				io=io->next;
		}
		else
		{	io=io->prev;	}
	}
}

int _io_extract( io_t *p )
{
	if( (p->prev->next=p->next)!=NULL )
		p->next->prev=p->prev;
	return(0);
}


io_t *io_add(io_t *io, io_t *prototyp )
{ io_t *new;
	if( (new=malloc(sizeof(io_t)))==NULL )	return(NULL);
	*new=*prototyp;
	new->evlist=NULL;
	if( _io_table_getfree(io,new) == 0xffffffff )
	{	free(new);	return(NULL);	}
	if( new->rd >= 0 )	io_setopen(io,new->rd);
	if( new->wr >= 0 )	io_setopen(io,new->wr);
	if( new->ex >= 0 )	io_setopen(io,new->ex);
	_io_add(io,new,new->prior);
	new->caller=io->id;
	if( new->func(new,IO_INIT,NULL)<0 )
	{	new->caller=new->id;
		io_delio(new);
		return(NULL);
	}
	return(new);
}

int io_delio( io_t *io )
{ io_t *p;
  int(*func)(io_t*,int,void*);
	if( io==NULL || io->func==NULL )	return(0);
	p=io;
	io=io->head;
	func=p->func;
	p->func=NULL;
	func(p,IO_DEL,NULL);
	if( p->rd >= 0 )	io_close(io,p->rd);
	if( p->wr >= 0 )	io_close(io,p->wr);
	if( p->ex >= 0 )	io_close(io,p->ex);
	p->rd=p->wr=p->ex=-1;
	p->inbuf=p->outbuf=p->ptr=p->ptr2=NULL;
	p->func=p->ufn=NULL;	/* Zombie */
	_delete_evlist(io);
	_io_table_setfree(io,p->id);
	return(0);
}

int io_chprior( io_t *io, int prior )
{ io_t *p;
	p=io;
	io=io->head;
	_io_extract(p);
	_io_add(io,p,prior);
	return(0);
}

int io_chrd( io_t *io, int rd )
{ io_t *p;
	p=io;
	io=io->head;
	if( p->rd >= 0 )
	{	if( p->wr==p->rd || p->ex==p->rd )	p->rd= -1;
		else					io_close(io,p->rd);
	}
	if( (p->rd=rd)>=0 )	io_setopen(io,p->rd);
	if( p->rd<0 && p->wr<0 && p->ex<0 )	io_callio(p,IO_CLOSED,NULL);
	return(0);
}

int io_chwr( io_t *io, int wr )
{ io_t *p;
	p=io;
	io=io->head;
	if( p->wr >= 0 )
	{	if( p->rd==p->wr || p->ex==p->wr )	p->wr= -1;
		else					io_close(io,p->wr);
	}
	if( (p->wr=wr)>=0 )	io_setopen(io,p->wr);
	if( p->rd<0 && p->wr<0 && p->ex<0 )	io_callio(p,IO_CLOSED,NULL);
	return(0);
}

int io_chex( io_t *io, int ex )
{ io_t *p;
	p=io;
	io=io->head;
	if( p->ex >= 0 )
	{	if( p->rd==p->ex || p->wr==p->ex )	p->ex= -1;
		else					io_close(io,p->ex);
	}
	if( (p->ex=ex)>=0 )	io_setopen(io,p->ex);
	if( p->rd<0 && p->wr<0 && p->ex<0 )	io_callio(p,IO_CLOSED,NULL);
	return(0);
}

int io_chtimer( io_t *io, unsigned long msec, unsigned long nmsec )
{
	io->msec=msec;
	io->nmsec=nmsec;
	return(0);
}

int io_exit(io_t *io, int retval)	/* retval!=0 */
{
	if( retval==0 || io==NULL )	return(-1);
	io=io->head;
	io->user1=retval;
	return(0);
}

int _io_do_call( io_t *p, int what, void *arg )
{
	if( p->func==NULL )	return(-1);
	if( p->func(p,what,arg)<0 )
	{	p->caller=p->id;
		io_delio(p);
		return(-1);
	}
	return(0);
}

int io_callio( io_t *io, int what, void *arg )
{
	if( what <= IO_LAST_SYSTEM )	return(-2);
	io->caller=0xffffffff;
	return(_io_do_call(io,what,arg));
}

int io_call( io_t *io, ioid_t id, int what, void *arg )
{ ioid_t cid;
	/* lock */
	cid=io->id;
	if( (io=io_find(io,id))==NULL )		return(-1);
	if( what!=IO_DEL && what <= IO_LAST_SYSTEM )	return(-2);
	io->caller=cid;
	if( what == IO_DEL )	return(io_delio(io));
	return(_io_do_call(io,what,arg));
}

long io_schedule( io_t *io, ioid_t id, long evid, unsigned long dtime, int what, void *arg, void(*ffree)(void*) )
{ evlist_t *e,*f,*fp;
  io_t *p;
	if( what!=IO_DEL && what <= IO_LAST_SYSTEM )	goto Err;
	if( id==IO_NULLID )	p=io;
	else if( (p=io_find(io,id))==NULL )		goto Err;
	if( p->func==NULL )				goto Err;
	if( ffree==STRING )
	{	if( (e=malloc(sizeof(evlist_t)+strlen(arg)+1))==NULL )
			goto Err;
		e->arg=&(e[1]);
		strcpy(e->arg,arg);
		e->free=NULL;
	}
	else if( ffree==INT )
	{	if( (e=malloc(sizeof(evlist_t)+sizeof(int)))==NULL )
			goto Err;
		e->arg=&(e[1]);
		*((int *)(e->arg))=*((int *)arg);
		e->free=NULL;
	}
	else
	{	if( (e=malloc(sizeof(evlist_t)))==NULL )	goto Err;
		e->arg=arg;
		e->free=ffree;
	}
	e->evid=iott(io->head)->last_evid++;
	if( (iott(io->head)->last_evid)<0 )
		iott(io->head)->last_evid=16;
	e->caller=io->id;
	e->dtime=dtime;
	e->what=what;
	f=p->evlist;	fp=NULL;
	if( evid>0 )
	{	while( f!=NULL && f->evid!=evid )
		{	fp=f;	f=f->next;	}
	}
	while( f!=NULL && f->dtime<=e->dtime )
	{	e->dtime-=f->dtime;
		fp=f;
		f=f->next;
	}
	if( fp==NULL )
		p->evlist=e;
	else
		fp->next=e;
	if( (e->next=f)!=NULL )
		f->dtime-=e->dtime;
	return(e->evid);

Err:	if( ffree!=NULL && ffree!=STRING && arg!=NULL )
		ffree(arg);
	return(0);
}

int io_do( io_t *io )
{ struct timeval tv;
  unsigned long gmsec,msec;
  int r,t,m,n;
  fd_set rd,wr,ex;
  io_t *p,*q;
  evlist_t *evlist;
  int i;
  unsigned long dtime;
	if( io==NULL || io!=io->head || io->prev!=NULL )
		return(-1);
	gettimeofday(&tv,NULL);
	gmsec=timeval2msec(tv);
	msec=gmsec;
	while( !(io->user1) )
	{	n=0;
		FD_ZERO(&rd);	FD_ZERO(&wr);	FD_ZERO(&ex);
		t=0; m= -1;
		for(p=io->next;p!=NULL;p=q)
		{	q=p->next;
			if( p->func==NULL )
			{	_io_extract(p);	/* remove zombie */
				free(p);
				continue;
			}
			n++;
//printf("io=%p rd=%d msec=%ld nmsec=%ld\n",p,p->rd,p->msec,p->nmsec);
			if(p->rd >= 0 && p->inbuf!=NULL)
			{	FD_SET(p->rd,&rd);
				if(p->rd > m)	m=p->rd;
			}
			if(p->wr >= 0 && p->outbuf!=NULL)
			{	FD_SET(p->wr,&wr);
				if(p->wr > m)	m=p->wr;
			}
			if(p->ex >= 0)
			{	FD_SET(p->ex,&ex);
				if(p->ex >m)	m=p->ex;
			}
			if(p->msec>0 && (!t || p->msec < msec))
			{	t=1;	msec=p->msec;	}
			if(p->evlist!=NULL && (!t || p->evlist->dtime < msec))
			{	t=1;	msec=p->evlist->dtime;	}
		}
		if( m==-1 && t==0 )	return(0);	/* nothing to do */
		tv.tv_sec= msec/1000;
		tv.tv_usec= (msec % 1000) * 1000;
//printf("io_do: io->user1=%d  n=%d\n",io->user1,n);
		r=select(m+1,&rd,&wr,&ex,(t?&tv:NULL));
		msec=gmsec;
		gettimeofday(&tv,NULL);
		gmsec=timeval2msec(tv);
		msec=gmsec-msec;
		if( r>=0 )
		{
		for(p=io->next;p!=NULL;p=p->next)
		{	if(p->msec>0)
			{	if( p->msec<=msec)
				{ dtime=msec-p->msec;
				  if( p->nmsec>0 )
				  { if( p->nmsec> dtime )
					p->msec=p->nmsec - dtime;
				    else	/* p->msec=1 ??? */
				    {	p->msec= dtime % p->nmsec ;
				    	i=(dtime-p->msec) / p->nmsec;
					p->msec= p->nmsec - p->msec;
					while( i>1 )
					{
			io_schedule(p,IO_NULLID,0,0,IO_TM_S,NULL,NULL);
					   i--;
					}
				    }
				  }
				  else	p->msec=0;
				  if(_io_do_call(p,IO_TM,NULL))	continue;
				}
				else
					p->msec-=msec;
			}
			if((evlist=p->evlist)!=NULL)
			{	dtime=msec;
				while(evlist!=NULL && evlist->dtime<=dtime)
				{ p->evlist=evlist->next;
				  dtime-=evlist->dtime;
				  p->caller=evlist->caller;
				  if( evlist->what==IO_DEL )	io_delio(p);
				  else _io_do_call(p,evlist->what,evlist->arg);
				  free(evlist);
				  evlist=p->evlist;
				}
				if(evlist!=NULL)
					evlist->dtime-=dtime;
			}
			if(p->rd >= 0 && FD_ISSET(p->rd,&rd))
				if(_io_do_call(p,IO_RD,NULL))	continue;
			if(p->wr >= 0 && FD_ISSET(p->wr,&wr))
				if(_io_do_call(p,IO_WR,NULL))	continue;
			if(p->ex >= 0 && FD_ISSET(p->ex,&ex))
				if(_io_do_call(p,IO_EX,NULL))	continue;
		}
		}
	}
	return(io->user1);
}


