/* Network library  V1.0  25.2.1997	*/
/* copyright (c)1997 by Marek Zelem	*/ 
/* e-mail: marek@crazy.elf.stuba.sk	*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <signal.h>
#include <termios.h>
#include <mmm/net_lib.h>

#define	iaddr_c2i(c)	(*(int *)(c))

pid_t	sig_exited_pid;
int	sig_signal[32];
int	sig_close,sig_shutdown;

/* Navratove hodnoty: 0-fatal error (not socket ... )
 *		      1-ok
 */


int net_verlib(char *ver)
{
	if(ver!=NULL)
		strcpy(ver,"Network library V2.2 (c)15.8.1997,14.10.98 by Marek Zelem");
	return(0x0022);
}

struct sockaddr *net_fill_sockaddrin( struct sockaddr_in *sockaddr, unsigned int ip, int port )
{
	sockaddr->sin_family = AF_INET;
	sockaddr->sin_addr.s_addr=ip;
	sockaddr->sin_port=htons(port);
	return((struct sockaddr *)sockaddr);
}

/* Odstraninie domeny, ak je taka ista ako ma lokalny stroj */
char *net_minimalname(char *name)
{ char pom[64];
  int i,j;
	  if( !getdomainname(pom+1,63) )
	  {	pom[0]='.';
	  	i=strlen(pom); j=strlen(name);
	  	if(j>i && !strcmp(name+j-i,pom) )
	  		name[j-i]=0;
	  }
	  return(name);
}


int net_ip2asci(unsigned int ip, char *asci)
{ unsigned char *s;
	s=(unsigned char *)&ip;
	sprintf(asci,"%u.%u.%u.%u",s[0],s[1],s[2],s[3]);
	return(1);
}

int net_name2ip(char *name, unsigned int *ip)
{ struct hostent *he;
	if(!name[0])
	{	*ip=(unsigned int)inet_addr ("127.0.0.1");
		return(1);
	}
	if( (*ip=(unsigned int)inet_addr(name))!=(unsigned int)(-1) )
		return(1);
	if( (he=gethostbyname(name))==NULL )
	    return(0);
	*ip=(unsigned int)iaddr_c2i(he->h_addr);
	return(1);
}

int net_ip2name(unsigned int ip, char *name)
{ struct hostent *he;

	if( (he=gethostbyaddr((unsigned char *)(&ip),4,AF_INET))!=NULL )
	{	strncpy(name,he->h_name,63);
		return(1);
	}
	else
		net_ip2asci(ip,name);
/*		strcpy(name,"???");	*/
	return(0);
	
}

char *net_maximalname( char *name )
{ struct hostent *he;
  unsigned int ip;

	if(!name[0])
		strcpy(name,"127.0.0.1");
//	if( name[0]=='0' && name[1]==0 )
//		strcpy(name,"127.0.0.1");
	if( (ip=(unsigned int)inet_addr (name))!=(unsigned int)(-1) )
	{	if( (he=gethostbyaddr((unsigned char *)(&ip),4,AF_INET))==NULL )
		{	net_ip2asci(ip,name);
			return(name);
		}
	}
	else if( (he=gethostbyname(name))==NULL )
	{
		return(name);
	}
	strncpy(name,he->h_name,63);
	return(name);
}

int net_serv2port(char *serv,char *proto)
{ struct servent *service;
  int r;
	setservent(1);
	if ((service=getservbyname(serv,proto))!=NULL)
		r=(int)ntohs(service->s_port);
	else
		r=0;
	endservent();
	return(r);
}

int net_text2port(char *text,char *proto)
{ int port;
	if( !text[0] )	return(23);
	if( (port=net_serv2port(text,proto))<=0 )
		port=atoi(text);
	return(port);
}

/* fcntl */

int net_s_noblock( int s )
{ int r;
	r=fcntl(s,F_GETFL,0);
	fcntl(s,F_SETFL,r|O_NONBLOCK|O_NDELAY);
	return(r);
}

int net_s_block( int s )
{ int r;
	r=fcntl(s,F_GETFL,0);
	fcntl(s,F_SETFL,r&~(O_NONBLOCK|O_NDELAY));
	return(r);
}

int net_s_set( int s, int fl )
{ int r;
	r=fcntl(s,F_GETFL,0);
	fcntl(s,F_SETFL,fl);
	return(r);
}

/* sietove funkcie */

int net_getpeer(int sock, char *name, unsigned int *ip, int *port )
{ struct sockaddr_in sinfo;
  int nlen;

   nlen=sizeof(struct sockaddr_in);
   if(!getpeername(sock,(struct sockaddr *)&sinfo,&nlen) )
	{
	  if( ip!=NULL )
	  	*ip=(unsigned int)sinfo.sin_addr.s_addr;
	  if( port!=NULL )
	  	*port=ntohs(sinfo.sin_port);
	  if( name!=NULL )
	  {
		net_ip2name((unsigned int)sinfo.sin_addr.s_addr,name);
	  }
	  return(1);
	}
   else
   	return(0);
}

int net_getme(int sock, char *name, unsigned int *ip, int *port )
{ struct sockaddr_in sinfo;
  int nlen;

   nlen=sizeof(struct sockaddr_in);
   if(!getsockname(sock,(struct sockaddr *)&sinfo,&nlen) )
	{
	  if( ip!=NULL )
	  	*ip=(unsigned int)sinfo.sin_addr.s_addr;
	  if( port!=NULL )
	  	*port=ntohs(sinfo.sin_port);
	  if( name!=NULL )
	  {
		net_ip2name((unsigned int)sinfo.sin_addr.s_addr,name);
	  }
	  return(1);
	}
   else
   	return(0);
}

/* Zistenie uzivatela */
int net_ident(unsigned int ip, int port, int myport, char *user)
{ int s;
  struct sockaddr_in iadr;
  char buf[1024];
  int max,pos,len,r;
  char stav[128],info[128];
	s=socket(PF_INET,SOCK_STREAM,0);
	iadr.sin_family = AF_INET;
	iadr.sin_addr.s_addr=ip;
	iadr.sin_port=htons(113);
	if( connect(s,(struct sockaddr*)&iadr,sizeof(struct sockaddr_in))>=0 )
	{
	sprintf(buf,"%d , %d\n",port,myport);
	write(s,buf,strlen(buf));
	pos=0;
	max=1024;
	while( (len=read(s,buf+pos,max))>0 )
		{ pos+=len; max-=len;
		  buf[pos]=0;
		  if( strchr(buf,'\n')!=NULL )	break;
		  if( strchr(buf,'\r')!=NULL )	break;
		}
	max=pos;
	buf[max]=0;
	user[0]=0;
	r=1;
	if( sscanf(buf,"%d , %d : %s : %s : %s",&pos,&pos,stav,info,user)!=5 )
	  if( sscanf(buf,"%d , %d : %s : %s :%s",&pos,&pos,stav,info,user)!=5 )
		r=0;
	if( r && !strncmp(stav,"USER",4))
		r=1;
	}
	else r=0;
	shutdown(s,2);
	close(s);
	return(r);
}

int net_sockident(int sock, char *host, char *name )
{ unsigned int ip;
  int port,myport;

	if( net_getpeer(sock,NULL, &ip, &port ) &&
	    net_getme(sock,NULL,NULL, &myport ) )
	{	if( host!=NULL )
			net_ip2name(ip,host);
		return( net_ident(ip,port,myport,name) );
	}
	return(0);
}

/* return socket if success or -1 if error */
int net_connect(unsigned int addr, int port)
{ int s;
  struct sockaddr_in saddr;
  
	s=socket(PF_INET,SOCK_STREAM,0);
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr=addr ;
	saddr.sin_port=htons(port);
	if( connect(s,(struct sockaddr*)&saddr,sizeof(struct sockaddr_in))>=0 )
		return(s);
	close(s);
	return(-1);
}

int net_bind(int port)
{ int s;
  struct sockaddr_in saddr;
  
	s=socket(PF_INET,SOCK_STREAM,0);
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr=INADDR_ANY;
	saddr.sin_port=htons(port);
	if( bind(s,(struct sockaddr*)&saddr,sizeof(struct sockaddr_in))>=0 )
	{	listen(s, 8);
		return(s);
	}
	close(s);
	return(-1);
}

int net_accept(int sock,char *name,unsigned int *ip,int *port)
{ struct sockaddr_in addr;
  int alen,r;
	alen=sizeof(struct sockaddr_in);
	if( (r=accept(sock,(struct sockaddr *)&addr,&alen))>=0 )
	{	if( ip!=NULL )
			*ip=(unsigned int)addr.sin_addr.s_addr;
		if( port!=NULL )
			*port=ntohs(addr.sin_port);
		if( name!=NULL )
		{
			net_ip2name((unsigned int)addr.sin_addr.s_addr,name);
		}
	}
	return(r);
}

int net_udpbind( int port )
{ int sock;
  struct sockaddr_in addr;

	sock=socket(PF_INET,SOCK_DGRAM,0);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= INADDR_ANY;
	addr.sin_port=htons(port);
	if( bind(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr_in))>=0 )
	{	listen(sock,8);
		return(sock);
	}
	close(sock);
	return(-1);
}

int net_udpsend(char *buf,int len,unsigned int ip,int port)
{ int sock;
  struct sockaddr_in addr;
	sock=socket(PF_INET,SOCK_DGRAM,0);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= ip;
	addr.sin_port=htons(port);
	if( sendto(sock,buf,len,0,(struct sockaddr*)&addr,sizeof(struct sockaddr_in))<0 )
	{	close(sock);
		return(-1);
	}
	return(sock);
}

int net_udprecv(int sock,char *buf,int maxlen,unsigned int *ip,int *port)
{ int len,fromlen;
  struct sockaddr_in addr;
	fromlen=sizeof(struct sockaddr_in);
	if( 
(len=recvfrom(sock,buf,maxlen,0/*MSG_WAITALL*/,(struct sockaddr*)&addr,&fromlen))
		>0 )
	{	if( ip!=NULL )
			*ip=addr.sin_addr.s_addr;
		if( port!=NULL )
			*port=ntohs(addr.sin_port);
	}
	return(len);
}

int net_udpconnect( unsigned int ip, int port )
{ struct sockaddr_in addr;
  int sock;
	sock=socket(PF_INET,SOCK_DGRAM,0);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= ip;
	addr.sin_port=htons(port);
	if( connect(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr_in))<0 )
	{	close(sock);
		return(-1);
	}
	return(sock);
}

int netnamed_connect(char *sockname)
{ int s;
  struct sockaddr_un saddr;
  
	s=socket(PF_UNIX,SOCK_STREAM,0);
	saddr.sun_family = AF_UNIX;
	strncpy(saddr.sun_path,sockname,sizeof(saddr.sun_path));
	if( connect(s,(struct sockaddr*)&saddr,sizeof(struct sockaddr_un))>=0 )
		return(s);
	close(s);
	return(-1);
}

int netnamed_bind(char *sockname)
{ int s;
  struct sockaddr_un saddr;
  
	s=socket(PF_UNIX,SOCK_STREAM,0);
	saddr.sun_family = AF_UNIX;
	strncpy(saddr.sun_path,sockname,sizeof(saddr.sun_path));
	if( bind(s,(struct sockaddr*)&saddr,sizeof(struct sockaddr_un))>=0 )
	{	listen(s, 8);
		return(s);
	}
	close(s);
	return(-1);
}

int netnamed_accept(int sock,char *remotename)
{ struct sockaddr_un addr;
  int alen,r;
	alen=sizeof(struct sockaddr_un);
	if( (r=accept(sock,(struct sockaddr *)&addr,&alen))>=0 )
	{	if( remotename!=NULL )
		{	strncpy(remotename,addr.sun_path,sizeof(addr.sun_path));
		}
	}
	return(r);
}

/* General function */

int net_gnrl_connect( int sock, unsigned int ip, int port )
{ struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= ip;
	addr.sin_port=htons(port);
	return( connect(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr_in)) );
}

int net_gnrl_bind( int sock, int port )
{ struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr= INADDR_ANY;
	addr.sin_port=htons(port);
	return( bind(sock,(struct sockaddr*)&addr,sizeof(struct sockaddr_in)) );
}

/* Extra specialne funkcie by Marek Zelem */

int net_GATEconnect(unsigned int Gaddr, int Gport, unsigned int addr, int port, int extra)
{ int s;
  unsigned char buf[32];
	memcpy(buf,"GATE\0x31\0x41\0x59\0x26",8);
	buf[8]=(unsigned char)((addr>>24)&0x0ff);
	buf[9]=(unsigned char)((addr>>16)&0x0ff);
	buf[10]=(unsigned char)((addr>>8)&0x0ff);
	buf[11]=(unsigned char)((addr)&0x0ff);
	buf[12]=(unsigned char)((port>>8)&0x0ff);
	buf[13]=(unsigned char)((port)&0x0ff);
	buf[14]=(unsigned char)((extra>>8)&0x0ff);
	buf[15]=(unsigned char)((extra)&0x0ff);
	if( (s=net_connect(Gaddr,Gport))<0 )	return(s);
	write(s,buf,16);
	return(s);
}

void net_sigchildexit(int signum)
{
	sig_exited_pid=waitpid(-1,NULL,WNOHANG);
}

void net_sigvoid(int signum)
{
	signal(signum,net_sigvoid);
}

void net_siginc(int signum)
{
	if( signum>=0 && signum <32 )	sig_signal[signum]++;
	signal(signum,net_siginc);
}

void net_sigclose(int signum)
{
	net_close(sig_close);
	signal(signum,net_sigclose);
}

void net_sigshutdown(int signum)
{ int r;
	r=shutdown(sig_shutdown,2);
	r=shutdown(sig_shutdown,2);
	signal(signum,net_sigshutdown);
}

void net_sigfork(int signum)
{
	if( fork()>0 )	exit(0);
	signal(signum,net_sigfork);
}

int net_close(int sock)
{
	shutdown(sock,2);
	close(sock);
	return(0);
}

/* Alokacia tty-ciek */

/* vrati fd pre pty alebo -1; in,out,io ak nie NULL, tak do nich zapise fd pre tty */
int net_ttyalloc_n( int *in, int *out, int *io, char *name )
{ char ttyname[16];
  int i,j,fp,ft;
	strcpy(ttyname,"/dev/ptyXX");
	for(i=0;(ttyname[8]="pqrstuvwxyzabcde"[i]);i++)
	  for(j=0;(ttyname[9]="0123456789abcdef"[j]);j++)
	{  if( (fp=open(ttyname,O_RDWR))>=0 )
	   {	ttyname[5]='t';
		ft=access(ttyname,R_OK|W_OK);
		if( io!=NULL && ft>=0 )
			ft=*io=open(ttyname,O_RDWR);
		if( in!=NULL && ft>=0 )
			ft=*in=open(ttyname,O_RDONLY);
		if( out!=NULL && ft>=0 )
			ft=*out=open(ttyname,O_WRONLY);
		if( ft>=0 )
		{	if( name!=NULL )
				strcpy(name,ttyname);
			return(fp);
		}
		if( in!=NULL && *in>=0 )	close(*in);
		if( out!=NULL && *out>=0 )	close(*out);
		if( io!=NULL && *io>=0 )	close(*io);
		close(fp);
	   	ttyname[5]='p';
	   }
	}
	return(-1);
}

int net_fileopen( char *filename, int *in, int *out, int *io )
{ int r;
	r=0;
	if( io!=NULL && r>=0 )
		r=*io=open(filename,O_RDWR);
	if( in!=NULL && r>=0 )
	{	if( (r=*in=open(filename,O_RDONLY))<0 )
		{	if( io!=NULL )	close(*io);
		}
	}
	if( out!=NULL && r>=0 )
	{	if( (r=*out=open(filename,O_WRONLY))<0 )
		{	if( io!=NULL )	close(*io);
			if( in!=NULL )	close(*in);
		}
	}
	return(r);
}

int net_ttyclean(int fd)
{ struct termios oldterm,newterm;
	tcgetattr(fd,&oldterm);
	newterm=oldterm;
	newterm.c_iflag=(IGNBRK|IGNPAR);
	newterm.c_oflag=(0);
	newterm.c_cflag=(CS8|CLOCAL);
	newterm.c_lflag=(0);
	tcsetattr(fd,TCSANOW,&newterm);
	return(0);
}

int net_ttyecho(int fd,int on)
{ struct termios oldterm,newterm;
  int r;
	tcgetattr(fd,&oldterm);
	newterm=oldterm;
	r=(newterm.c_lflag&ECHO?1:0);
	if( on )
		newterm.c_lflag|=(ECHO);
	else
		newterm.c_lflag&=~(ECHO);
	tcsetattr(fd,TCSANOW,&newterm);
	return(r);
}

/* Kolosalne funkcie */

pid_t net_tcpserver(int sock, void(*child)(int))
{ int sc;
  pid_t pid;
	if( (sc=net_accept(sock,NULL,NULL,NULL))<0 )
		return(-1);
	if( (pid=fork())==0 )
	{	child(sc);
		net_close(sc);
		exit(0);
	}
	else if( pid<0 )
		return(-2);
	return(pid);
}

int my_copy(int from, int to )
{ char buf[2048];
  int len;
	while( (len=read(from,buf,2048))>0 )
		write(to,buf,len);
	return(len);
}

int net_printf(int sock,char *fmt, ...)
{ va_list ap;
  int len;
  char buf[4096];
	va_start(ap,fmt);
	len=vsnprintf(buf,4096,fmt,ap);
	va_end(ap);
	return(write(sock,buf,len));
}

int net_gets(int sock,char *buf,int maxlen)
{ int i,l;
	i= -1; buf[0]=0; l=1;
	while(l==1 && i<maxlen && buf[i]!='\r' )
	{	if( (l=read(sock,buf+i+1,1))==1 && buf[i+1]!='\n' ) i++;
	}
	buf[i]=0;
	return(i);
}

int net_getln(int sock,char *buf,int maxlen)
{ int i;
	i=0;
	while( i<(maxlen-1) && read(sock,buf+i,1)==1 )
	{	i++;
		if( buf[i-1]=='\n' )	break;
	}
	buf[i]=0;
	return(i);
}

int net_cross(int s1,int s2)
{ fd_set rd;
  char buf[128];
  int len,er,n;
	er=0;
	while(!er)
	{
	FD_ZERO(&rd);
	FD_SET(s1,&rd);
	FD_SET(s2,&rd);
	n=s2; if(s1>n) n=s1;
	n++;
	select(n,&rd,NULL,NULL,NULL);
	if( FD_ISSET(s1,&rd) )
	{
		if( (len=read(s1,buf,128))>0 )
			write(s2,buf,len);
		else
			er=1;
	}
	if( FD_ISSET(s2,&rd) )
	{
		if( (len=read(s2,buf,128))>0 )
			write(s1,buf,len);
		else
			er=1;
	}
	}
	return(0);
}

int net_tostdio(int s, int in, int out)
{ fd_set rd;
  char buf[128];
  int len,er,n;
	er=0;
	while(!er)
	{
	FD_ZERO(&rd);
	FD_SET(s,&rd);
	FD_SET(in,&rd);
	n=s; if(in>n) n=in;
	n++;
	select(n,&rd,NULL,NULL,NULL);
	if( FD_ISSET(s,&rd) )
	{
		if( (len=read(s,buf,128))>0 )
			write(out,buf,len);
		else
			er=1;
	}
	if( FD_ISSET(in,&rd) )
	{
		if( (len=read(in,buf,128))>0 )
			write(s,buf,len);
		else
			er=1;
	}
	}
	return(0);
}

int net_crossfunc(int s1,int s2,int(*f12)(char *,int) ,int(*f21)(char *,int) )
{ fd_set rd;
  char buf[128];
  int len,er,n;
 	er=0;
 	while(!er)
 	{
 	FD_ZERO(&rd);
 	FD_SET(s1,&rd);
 	FD_SET(s2,&rd);
 	n=s2; if(s1>n) n=s1;
 	n++;
 	select(n,&rd,NULL,NULL,NULL);
 	if( FD_ISSET(s1,&rd) )
 	{
 		if( (len=read(s1,buf,128))>0 )
 		{	if( (len=f12(buf,len))>0 )
 				write(s2,buf,len);
 		}
 		else
 			er=1;
 	}
 	if( FD_ISSET(s2,&rd) )
 	{
 		if( (len=read(s2,buf,128))>0 )
 		{	if( (len=f21(buf,len))>0 )
 				write(s1,buf,len);
 		}
 		else
 			er=1;
 	}
 	}
 	return(0);
}

int net_comm(comm_t *tab,int n )
{ fd_set rd;
  int er,i,m;
	er=0;
	while(!er)
	{
	m=0;
	FD_ZERO(&rd);
	for(i=0;i<n;i++)
	{	FD_SET(tab[i].fd,&rd);
		if( tab[i].fd>m )	m=tab[i].fd;
	}
	m++;
	select(m,&rd,NULL,NULL,NULL);
	er=0;
	for(i=0;i<n;i++)
	{ if( FD_ISSET(tab[i].fd,&rd) )
		er+=tab[i].func(tab[i].fd,&(tab[i].p1),&(tab[i].p2));
	}

	}
	return(er);
}

int net_mccp(int fd, int *p1, int *p2)	/*modul k net_comm  p1-fd na zapis*/
{ int len;
  char buf[128];
	if( (len=read(fd,buf,128))>0 )
	{	write(*p1,buf,len);
		return(0);
	}
	*p1= -1;
	return(1);
}
