/* Network app library  V1.0  15.8.1997	*/
/* copyright (c)1997 by Marek Zelem	*/ 
/* e-mail: marek@crazy.elf.stuba.sk	*/

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <termios.h>
#include <pwd.h>
#define	osockaddr	sockaddr
#include <protocols/talkd.h>
#include <errno.h>
#include <mmm/net_lib.h>
#include <mmm/netap_lib.h>

int netap_verlib(char *ver)
{
	if(ver!=NULL)
		strcpy(ver,"Network App library V1.0 (c)15.8.1997 by Marek Zelem");
	return(0x0010);
}


unsigned int netap_host_ip( void )
{ unsigned int ip;
  char host[64];
	gethostname(host,64);
	host[63]=0;
	if( !net_name2ip(host,&ip) )
		return(0);
	return(ip);
}

char *netap_uid2login( int uid )
{ struct passwd *p;
  static char name[16];
	if( (p=getpwuid(uid))!=NULL )
	{  strncpy(name,p->pw_name,16);
	   name[16]=0;
	}
	else
	{  sprintf(name,"#%d",uid);
	}
	return(name);
}

int netap_login2uid( const char *loginname )
{ struct passwd *p;
	if( (p=getpwnam(loginname))!=NULL )
	{  return((int)(p->pw_uid));
	}
	return(-1);
}

char *netap_username( void )
{ static char name[32];
  char *p;

  	name[0]=0;
#ifdef linux
	cuserid(name);
	if( name[0] );
	else
#endif linux
	     if( (p=getlogin())!=NULL )
	{	strncpy(name,p,16);
		name[16]=0;
	}
	else
	{	strncpy(name,netap_uid2login(getuid()),16);
		name[16]=0;
	}
	return(name);
}



/* ---- Komplexne funkcie ---- */

/* netap_tell
 *	addr	- adresa vzdialeneho stroja
 *	port	- cislo portu, alebo 0 ak standardny
 *	fromname- meno odosielatela
 *	toname	- meno adresata
 *	text	- text spravy.
 *	out	- Pri vystupe nahradena textom odpovede. (max 1024)
 * RETURN VALUE
 *	0	- error - can't connect
 *	1	- o.k.
 */
int netap_tell(unsigned int addr, int port, char *fromname, char *toname, char *utty, char *text, char *out)
{ int s;
  int len;
  
	if( port<=0 )
	{  if( (port=net_serv2port("tell","tcp"))<=0 )	port=4224;
	}
	if( (s=net_connect(addr,port))<0 )	return(0);
	net_printf(s,"%s:%s:%s:%s\n",fromname,toname,utty,text);
	if( (len=read(s,out,1023))>0 )
	{	out[len]=0;
	}
	net_close(s);
	return(1);
}

/* talk library	*/

int netap_talk_open( talk_t *talk, unsigned int ip, int port, char *myname )
{ int i;
	if( port<=0 )
	{  if( (port=net_serv2port("ntalk","udp"))<=0 )	port=518;
	}
	if( ip==0 )
	{  ip=netap_host_ip();
	}
	talk->myserver_ip=ip;
	talk->myserver_port=port;
	if( (talk->ctl_sock=socket(PF_INET,SOCK_DGRAM,0))<0 )
		return(-1);
	net_gnrl_bind(talk->ctl_sock,0);
	net_getme(talk->ctl_sock,NULL,NULL,&(talk->ctl_port));
	talk->ctl_ip=netap_host_ip();
	if( myname!=NULL )
		strncpy(talk->l_name,myname,NAME_SIZE);
	else
		strncpy(talk->l_name,netap_username(),NAME_SIZE);
	talk->repeat_invite=RING_WAIT;
	talk->repeat_announce=20;
	for(i=0;i<MAX_INVITES;i++)	talk->invite[i].sock= -1;
	talk->invites=0;
	return(0);
}

int netap_talk_close( talk_t *talk )
{ int i;
	for(i=0;i<MAX_INVITES;i++)
	{  if( talk->invite[i].sock>=0 )
	   {	netap_talk_request(DELETE,talk,i);
		netap_talk_receive(talk,NULL,NULL);
	   }
	}
	net_close(talk->ctl_sock);
	return(0);
}

int netap_talk_i_new( talk_t *talk, char *r_name, char *r_tty )
{ int i;
	for(i=0;i<MAX_INVITES && talk->invite[i].sock>=0;i++);
	if(i>=MAX_INVITES)	return(-1);
	if( (talk->invite[i].sock=net_bind(0))<0 )	return(-2);
	net_getme(talk->invite[i].sock,NULL,NULL,&(talk->invite[i].sock_port));
	talk->invite[i].sock_ip=netap_host_ip();
	talk->invite[i].time=0;
	talk->invite[i].announce_time=0;
	strncpy(talk->invite[i].r_name,r_name,NAME_SIZE);
	strncpy(talk->invite[i].r_tty,r_tty,TTY_SIZE);
	talk->invites++;
	return(i);
}

int netap_talk_i_find( talk_t *talk, char *r_name, char *r_tty )
{ int i;
	for(	i=0;
		i<MAX_INVITES 
		&& (talk->invite[i].sock<0 
			|| strncmp(talk->invite[i].r_name,r_name,NAME_SIZE)
			|| strncmp(talk->invite[i].r_tty,r_tty,TTY_SIZE) ) ;
		i++);
	if(i>=MAX_INVITES)	return(-1);
	return(i);
}

int netap_talk_i_del( talk_t *talk, int n )
{
	if( talk->invite[n].sock<0 )
		return(0);
	net_close(talk->invite[n].sock);
	talk->invite[n].sock= -1;
	talk->invites--;
	return(1);
}


/* netap_talk_invite   - interna funkcia na zaslanie invitation
 *   talk	- struktura talk, otvorena pomocou natap_talk_open
 *   ip		- nepovinna ip adresa ( ak 0 tak vyberie sam )
 *   port	- nepovinny port (ak 0 tak vyberie sam )
 *   toname	- meno volaneho
 *   to_tty	- tty volaneho
 * Return value	- <0 ak error. num ak OK
 */
int netap_talk_invite( talk_t *talk, unsigned int ip, int port, char *toname, char *to_tty )
{ int num,r;
	if( port<=0 )
		port=talk->myserver_port;
	if( ip==0 )
		ip=talk->myserver_ip;
	if( (num=netap_talk_i_new(talk,toname,to_tty))<0 )
		return(-1);
	talk->invite[num].server_ip=ip;
	talk->invite[num].server_port=port;
	talk->invite[num].time=time(NULL);
	if( (r=netap_talk_request(LEAVE_INVITE,talk,num))<0 )
	{	netap_talk_i_del(talk,num);
		return(-1);
	}
	return(r);
}

/* netap_talk_announce   - interna funkcia na zaslanie announce
 *   talk	- struktura talk, otvorena pomocou natap_talk_open
 *   num	- num of invitation
 *   ip		- nepovinna ip adresa ( ak 0 tak vyberie sam )
 *   port	- nepovinny port (ak 0 tak vyberie sam )
 * Return value	- <0 ak error. num ak OK
 */
int netap_talk_announce( talk_t *talk, int num, unsigned int ip, int port )
{
	if( num<0 || talk->invite[num].sock<0 )
		return(-1);
	if( port<=0 )
	{	if( (port=net_serv2port("ntalk","udp"))<=0 )	port=518;
	}
	if( ip==0 )
		ip=netap_host_ip();
	talk->invite[num].announce_ip=ip;
	talk->invite[num].announce_port=port;
	talk->invite[num].announce_time=time(NULL);
	talk->invite[num].announce_id_num=5;
	talk->invite[num].time=time(NULL);
	return(netap_talk_request(ANNOUNCE,talk,num));
}

/* netap_talk_lookup   - interna funkcia na zaslanie lookup
 *   type	- typ reqestu. (LOOK_UP, ANNOUNCE)
 *   talk	- struktura talk, otvorena pomocou natap_talk_open
 *   ip		- nepovinna ip adresa ( ak 0 tak vyberie sam )
 *   port	- nepovinny port (ak 0 tak vyberie sam )
 *   toname	- meno volaneho
 *   to_tty	- tty volaneho
 * Return value	- <0 ak error. 0 alebo num ak OK
 * PRIKLADY POUZITIA:
 *   (LOOK_UP,talk,ip,0,toname,to_tty)		- <0 - err | 0 - OK
 *   (ANNOUNCE,talk,ip,0,toname,to_tty)		- <0 - err | 0 -OK
 */
int netap_talk_lookup( int type, talk_t *talk, unsigned int ip, int port, char *toname, char *to_tty )
{ CTL_MSG ctl;
  struct sockaddr_in sad;
	if( port<=0 )
	{	if( (port=net_serv2port("ntalk","udp"))<=0 )	port=518;
	}
	if( ip==0 )
		ip=netap_host_ip();
	ctl.vers=TALK_VERSION;
	ctl.type=type;
	ctl.id_num=htonl(0);
	sad.sin_family = htons(AF_INET);
	memcpy(&(ctl.addr),&sad,sizeof(sad));
	sad.sin_addr.s_addr=talk->ctl_ip;
	sad.sin_port=htons(talk->ctl_port);
	memcpy(&(ctl.ctl_addr),&sad,sizeof(sad));
	ctl.pid=htonl(getpid());
	strncpy(ctl.l_name,talk->l_name,NAME_SIZE);
	strncpy(ctl.r_name,toname,NAME_SIZE);
	strncpy(ctl.r_tty,to_tty,TTY_SIZE);
	net_fill_sockaddrin(&sad,ip,port);
        talk->ctl_buf = ctl;
        talk->sad = sad;
        talk->num = -1;
	return(0);
}

/* netap_talk_request   - interna funkcia na zaslanie requestu
 *   type	- typ reqestu. (LOOK_UP, LEAVE_INVITE, DELETE, ANNOUNCE)
 *   talk	- struktura talk, otvorena pomocou natap_talk_open
 *   num	- pouzije ip a port, r_name a r_tty z invitation table[num]
 * Return value	- <0 ak error. 0 alebo num ak OK
 * PRIKLADY POUZITIA:
 *   (LOOK_UP,talk,num)		- <0 - err | num - OK
 *   (LEAVE_INVITE,talk,num)	- <0 - err | num - OK ; opakovanie invitation
 *   (DELETE,talk,num)		- <0 - err | num - OK
 *   (ANNOUNCE,talk,num)	- <0 - err | num -OK
 */
int netap_talk_request( int type, talk_t *talk, int num )
{ CTL_MSG ctl;
  struct sockaddr_in sad;
  unsigned int ip;
  int port;
	if( num<0 || talk->invite[num].sock<0 )
		return(-1);
	if( type==ANNOUNCE || type==LOOK_UP )
	{  port=talk->invite[num].announce_port;
	   ip=talk->invite[num].announce_ip;
	}
	else
	{  port=talk->invite[num].server_port;
	   ip=talk->invite[num].server_ip;
	}
	ctl.vers=TALK_VERSION;
	ctl.type=type;
	ctl.id_num=htonl(talk->invite[num].id_num);
	sad.sin_family = htons(AF_INET);
	sad.sin_addr.s_addr=talk->invite[num].sock_ip;
	sad.sin_port=htons(talk->invite[num].sock_port);
	if(type==LEAVE_INVITE)
	{	talk->invite[num].time=time(NULL);
	}
	else if(type==DELETE)
	{	netap_talk_i_del(talk,num);
	}
	else if(type==ANNOUNCE)
	{	talk->invite[num].announce_time=time(NULL);
		talk->invite[num].announce_id_num+=1;
		ctl.id_num=htonl(talk->invite[num].announce_id_num);
	}
	else
	{	ctl.id_num=htonl(0);
	}
	memcpy(&(ctl.addr),&sad,sizeof(sad));
	sad.sin_addr.s_addr=talk->ctl_ip;
	sad.sin_port=htons(talk->ctl_port);
	memcpy(&(ctl.ctl_addr),&sad,sizeof(sad));
	ctl.pid=htonl(getpid());
	strncpy(ctl.l_name,talk->l_name,NAME_SIZE);
	strncpy(ctl.r_name,talk->invite[num].r_name,NAME_SIZE);
	strncpy(ctl.r_tty,talk->invite[num].r_tty,TTY_SIZE);
	net_fill_sockaddrin(&sad,ip,port);
	talk->ctl_buf = ctl;
	talk->sad = sad;
	talk->num = num;
	return(num);
}

/* netap_talk_receive	- interna funkcia na prijatie odpovede
 *   talk	- struktura talk, otvorena pomocou natap_talk_open
 *   ip		- nepovinny parameter ak answer=0 potom sem zapise ip z addr
 *   port	- nepovinny parameter ak answer=0 potom sem zapise port z addr
 * Return value:
 *   <0 error 0-ok 1-not here
 */
int netap_talk_receive( talk_t *talk, unsigned int *ip, int *port )
{ CTL_RESPONSE ctl_resp;
  int r,num;
  fd_set fds;
  struct timeval tv;
  struct sockaddr_in	*sin;
	num=talk->num;
   do {
	do {	
        	if( sendto(talk->ctl_sock,&(talk->ctl_buf),sizeof(talk->ctl_buf),0,(struct sockaddr*)&(talk->sad),sizeof(struct sockaddr_in))!=sizeof(talk->ctl_buf) )
			return(-1);
		FD_ZERO(&fds);
		FD_SET(talk->ctl_sock,&fds);
		tv.tv_sec=4;	tv.tv_usec=0;
		r=select(talk->ctl_sock+1,&fds,NULL,NULL,&tv);
	   } while( r<=0 );
	if(recv(talk->ctl_sock,&ctl_resp,sizeof(ctl_resp),0)!=sizeof(ctl_resp))
		return(-1);
	if( ctl_resp.vers!=TALK_VERSION )
		return(-2);
      } while( ctl_resp.type!=talk->ctl_buf.type );
	r=ctl_resp.answer;
	if( r>1 )
		return(-10-r);
	if( num>=0 && talk->invite[num].sock>=0 )
	{  if( talk->ctl_buf.type==ANNOUNCE )
		talk->invite[num].announce_id_num=ntohl(ctl_resp.id_num);
	   else
		talk->invite[num].id_num=ntohl(ctl_resp.id_num);
	}
	sin=(struct sockaddr_in *)&(ctl_resp.addr);
	if( r==0 )
	{	if( sin->sin_family!=htons(AF_INET) )
		{  if(ip!=NULL)		*ip=0;
		   if(port!=NULL)	*port=0;
		   return(r);
		}
		if( ip!=NULL )
			*ip=sin->sin_addr.s_addr;
		if( port!=NULL )
			*port=ntohs(sin->sin_port);
	}
	return(r);
}

int netap_talk_talk( talk_t *talk, unsigned int ip, int port, char *toname, char *to_tty, int(*func)(int,char*,long), long param )
{ int r,sock,num;
  unsigned int nip;
  int nport;
	if( netap_talk_lookup(LOOK_UP,talk,ip,port,toname,to_tty )<0 )
		return(-1);
	if( (r=netap_talk_receive(talk,&nip,&nport))<0 )
		return(-1);
	if( r==0 )
	{	if( (sock=net_connect(nip,nport))>=0 )
			return( func(sock,toname,param) );
	}
	if( (num=netap_talk_invite(talk,0,0,toname,to_tty))<0 )
		return(-2);
	if( (r=netap_talk_receive(talk,NULL,NULL))<0 )
	{	netap_talk_request(DELETE,talk,num);
		netap_talk_receive(talk,NULL,NULL);
		return(-100+r);
	}
	if( netap_talk_lookup(LOOK_UP,talk,ip,port,toname,to_tty )<0 )
		return(-10);
	if( (r=netap_talk_receive(talk,&nip,&nport))<0 )
		return(-10);
	if( r==0 )
	{	if( (sock=net_connect(nip,nport))>=0 )
		{  netap_talk_request(DELETE,talk,num);
		   netap_talk_receive(talk,NULL,NULL);
		   return( func(sock,toname,param) );
		}
	}
	talk->invite[num].param=param;
	talk->invite[num].func=func;
	netap_talk_announce(talk,num,ip,port);
	netap_talk_receive(talk,NULL,NULL);
	return(0);
}


int netap_talk_comm( talk_t *talk, comm_t *tab, int n )
{ fd_set rd;
  int er,i,m,sock;
  time_t t;
  struct timeval tv;
  CTL_RESPONSE ctl_resp;
	er=0;
	while(!er)
	{
	m=0;
	FD_ZERO(&rd);
	FD_SET(talk->ctl_sock,&rd);
	m=talk->ctl_sock;
	if(talk->invites<=0 && (tab==NULL || n<=0) )
		return(0);	/* Nothing to do */
	t=time(NULL);
	for(i=0;i<MAX_INVITES;i++)
	{  if(talk->invite[i].sock>=0)
	   {  if( t-(talk->invite[i].time) > talk->repeat_invite )
	      {  if( netap_talk_request(LEAVE_INVITE,talk,i)>=0 )
		 	netap_talk_receive(talk,NULL,NULL);
	      }
	      if( talk->invite[i].announce_time>0 && t-(talk->invite[i].announce_time) > talk->repeat_announce )
	      {  if( netap_talk_request(ANNOUNCE,talk,i)>=0 )
			netap_talk_receive(talk,NULL,NULL);
	      }
	      FD_SET(talk->invite[i].sock,&rd);
	      if( talk->invite[i].sock > m )	m=talk->invite[i].sock;
	   }
	}
	if( tab!=NULL && n>0 )
	{  for(i=0;i<n;i++)
	   {	FD_SET(tab[i].fd,&rd);
		if( tab[i].fd>m )	m=tab[i].fd;
	   }
	}
	m++;
	tv.tv_sec=5; tv.tv_usec=0;
	er=0;
	if( select(m,&rd,NULL,NULL,&tv) )
	{  if( FD_ISSET(talk->ctl_sock,&rd) )
	   {	recv(talk->ctl_sock,&ctl_resp,sizeof(ctl_resp),0);
	   }
	   for(i=0;i<MAX_INVITES;i++)
	   {  if(talk->invite[i].sock>=0)
	      {  if( FD_ISSET(talk->invite[i].sock,&rd) )
	         {	sock=net_accept(talk->invite[i].sock,NULL,NULL,NULL);
	           if( netap_talk_request(DELETE,talk,i)>=0 )
			netap_talk_receive(talk,NULL,NULL);
		   er+=talk->invite[i].func(sock,talk->invite[i].r_name,talk->invite[i].param);
	         }
	      }
	   }
	   if( tab!=NULL && n>0 )
	   {  for(i=0;i<n;i++)
	      { if( FD_ISSET(tab[i].fd,&rd) )
		   er+=tab[i].func(tab[i].fd,&(tab[i].p1),&(tab[i].p2));
	      }
	   }
	}
	}
	return(er);
}




/* -------------------------------- */

int pok_func(int sock, char *name, long par)
{
	printf("Connected with %s\n",name);
	fflush(stdout);
	net_tostdio(sock,0,1);
	return(5);
}

#ifdef DEBUG
int main( int argc, char *argv[] )
{ unsigned int ip;
  char text[1024];
  talk_t talk;
  int r;
	net_name2ip(argv[2],&ip);
/*
	printf("%d\n",netap_tell(ip,0,netap_username(),argv[1],"","Ahoj",text));
	printf("%s\n",text);
*/

	if( netap_talk_open(&talk,0,0,NULL)<0 )
		return(-1);
	r=netap_talk_talk(&talk,ip,0,argv[1],"",pok_func,0);
	printf("talk: %d\n",r);
	netap_talk_comm(&talk,NULL,0);
	printf("END\n");
	return(0);
}
#endif DEBUG

