/* Network app library  V1.0  15.8.1997	*/
/* copyright (c)1997 by Marek Zelem	*/ 
/* e-mail: marek@crazy.elf.stuba.sk	*/

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <termios.h>
#include <pwd.h>
#define	osockaddr	sockaddr
#include <protocols/talkd.h>
#include <errno.h>
#include <mmm/net_lib.h>
#include <mmm/netap_lib.h>

int netap_verlib(char *ver)
{
	if(ver!=NULL)
		strcpy(ver,"Network App library V1.0 (c)15.8.1997 by Marek Zelem");
	return(0x0010);
}


unsigned int netap_host_ip( void )
{ unsignet int ip;
  host[64];
	gethostname(host,64);
	host[63]=0;
	if( !net_name2ip(host,&ip) )
		return(0);
	return(ip);
}

char *netap_uid2login( int uid )
{ struct passwd *p;
  static char name[16];
	if( (p=getpwuid(uid))!=NULL )
	{  strncpy(name,p->pw_name,16);
	   name[16]=0;
	}
	else
	{  sprintf(name,"#%s",uid);
	}
	return(name);
}

int netap_login2uid( const char *loginname )
{ struct passwd *p;
	if( (p=getpwnam(loginname))!=NULL )
	{  return((int)(p->pw_uid));
	}
	return(-1);
}

char *netap_username( void )
{ static char name[32];
  char *p;

	cuserid(name);
	if( name[0] );
	else if( (p=getlogin())!=NULL )
	{	strncpy(name,p,16);
		name[16]=0;
	}
	else
	{	strncpy(name,netap_uid2login(getuid()),16);
		name[16]=0;
	}
	return(name);
}



/* ---- Komplexne funkcie ---- */

/* netap_tell
 *	addr	- adresa vzdialeneho stroja
 *	port	- cislo portu, alebo 0 ak standardny
 *	fromname- meno odosielatela
 *	toname	- meno adresata
 *	text	- text spravy.
 *	out	- Pri vystupe nahradena textom odpovede. (max 1024)
 * RETURN VALUE
 *	0	- error - can't connect
 *	1	- o.k.
 */
int netap_tell(unsigned int addr, int port, char *fromname, char *toname, char *utty, char *text, char *out)
{ int s;
  struct sockaddr_in saddr;
  int len;
  
	if( port<=0 )
	{  if( (port=net_serv2port("tell","tcp"))<=0 )	port=4224;
	}
	if( (s=net_connect(addr,port))<0 )	return(0);
	net_printf(s,"%s:%s:%s:%s\n",fromname,toname,utty,text);
	if( (len=read(s,out,1023))>0 )
	{	out[len]=0;
	}
	net_close(s);
	return(1);
}

/* talk library	*/

int netap_talk_ctlconnect( unsigned int ip, int port )
{ int sock;

	if( port<=0 )
	{  if( (port=net_serv2port("ntalk","udp"))<=0 )	port=518;
	}
	if( (sock=net_udpconnect(ip,port))<0 )	return(-1);
	return(sock);
}

int netap_talk_lookup( int sock, long id_num, char *fromname, char *toname, char *to_tty )
{ CTL_MSG ctl;
  struct sockaddr_in sad;
  unsigned int myip;
  int myport;
	net_getme(sock,NULL,&myip,&myport);
	ctl.vers=TALK_VERSION;
	ctl.type=LOOK_UP;
	ctl.id_num=htonl(0);
	sad.sin_family = htons(AF_INET);
	memcpy(&(ctl.addr),&sad,sizeof(sad));
	sad.sin_addr.s_addr=myip;
	sad.sin_port=htons(myport);
	memcpy(&(ctl.ctl_addr),&sad,sizeof(sad));
	ctl.pid=htonl(getpid());
	strncpy(ctl.l_name,fromname,NAME_SIZE);
	strncpy(ctl.r_name,toname,NAME_SIZE);
	strncpy(ctl.r_tty,to_tty,TTY_SIZE);
	if( send(sock,&ctl,sizeof(ctl),0)!=sizeof(ctl) )
	{	net_close(sock);
		return(-1);
	}
	return(sock);
}

int netap_talk_wait4caller( int sock, long *id_num, struct sockaddr_in *addr )
{
	return( netap_talk_wait4answer(sock,LOOK_UP,id_num,addr) );
}
int netap_talk_wait4answer( int sock, int type, long *id_num, struct sockaddr_in *addr )
{ CTL_RESPONSE ctl_resp;
  int r;
	if( recv(sock,&ctl_resp,sizeof(ctl_resp),0)!=sizeof(ctl_resp) )
		return(-1);
	if( ctl_resp.vers!=TALK_VERSION || ctl_resp.type!=type )
		return(-2);
	r=ctl_resp.answer;
	if( r>1 )
		return(-10-r);
	if( r==0 )
	{	if( addr!=NULL )
		{  memcpy(addr,&(ctl_resp.addr),sizeof(struct sockaddr_in));
		   addr->sin_family=ntohs(addr->sin_family);
		}
		if( id_num!=NULL )
			*id_num=ntohl(ctl_resp.id_num);
	}
	return(r);
}

int netap_talk_announce( int sock, long id_num, char *fromname, char *toname, char *to_tty )
{ CTL_MSG ctl;
  struct sockaddr_in sad;
  unsigned int myip;
  int myport;
	net_getme(sock,NULL,&myip,&myport);
	ctl.vers=TALK_VERSION;
	ctl.type=ANNOUNCE;
	ctl.id_num=htonl(0);
	sad.sin_family = htons(AF_INET);
	memcpy(&(ctl.addr),&sad,sizeof(sad));
	sad.sin_addr.s_addr=myip;
	sad.sin_port=htons(myport);
	memcpy(&(ctl.ctl_addr),&sad,sizeof(sad));
	ctl.pid=htonl(getpid());
	strncpy(ctl.l_name,fromname,NAME_SIZE);
	strncpy(ctl.r_name,toname,NAME_SIZE);
	strncpy(ctl.r_tty,to_tty,TTY_SIZE);
	if( send(sock,&ctl,sizeof(ctl),0)!=sizeof(ctl) )
	{	net_close(sock);
		return(-1);
	}
	return(sock);
}

int netap_talk_invite( int sock, long id_num, char *fromname, char *toname, char *to_tty )
{ CTL_MSG ctl;
  struct sockaddr_in sad;
  unsigned int myip;
  int myport;
	net_getme(sock,NULL,&myip,&myport);
	ctl.vers=TALK_VERSION;
	ctl.type=LEAVE_INVITE;
	ctl.id_num=htonl(0);
	sad.sin_family = htons(AF_INET);
	sad.sin_addr.s_addr=myip;
	sad.sin_port=htons(32132);
	memcpy(&(ctl.addr),&sad,sizeof(sad));
	sad.sin_addr.s_addr=myip;
	sad.sin_port=htons(myport);
	memcpy(&(ctl.ctl_addr),&sad,sizeof(sad));
	ctl.pid=htonl(getpid());
	strncpy(ctl.l_name,fromname,NAME_SIZE);
	strncpy(ctl.r_name,toname,NAME_SIZE);
	strncpy(ctl.r_tty,to_tty,TTY_SIZE);
	if( send(sock,&ctl,sizeof(ctl),0)!=sizeof(ctl) )
	{	net_close(sock);
		return(-1);
	}
	return(sock);
}

/* -------------------------------- */

int main( int argc, char *argv[] )
{ int sock;
  unsigned int ip;
  char text[1024];
  long id_num;
  struct sockaddr_in addr;
  int r;
	net_name2ip(argv[2],&ip);
/*
	printf("%d\n",netap_tell(ip,0,netap_username(),argv[1],"","Ahoj",text));
	printf("%s\n",text);
*/
	if( (sock=netap_talk_ctlconnect(ip,0))<0 )
		return(-1);
	if( netap_talk_lookup(sock,0,netap_username(),argv[1],"")<0 )
	{	printf("err\n");
		return(-1);
	}
	r=netap_talk_wait4answer(sock,LOOK_UP,&id_num,&addr);
	printf("answer: %d\nid_num: %d\nport: %d\n",r,id_num,ntohs(addr.sin_port));
	if( netap_talk_invite(sock,netap_username(),argv[1],"")<0 )
	{	printf("announce err\n");
		return(-1);
	}
/*
	if( netap_talk_lookup(sock,netap_username(),argv[1],"")<0 )
	{	printf("err\n");
		return(-1);
	}
*/
	r=netap_talk_wait4answer(sock,LEAVE_INVITE,&id_num,&addr);
	printf("answer: %d\nid_num: %d\nport: %d\n",r,id_num,ntohs(addr.sin_port));
	r=netap_talk_wait4caller(sock,&id_num,&addr);
	printf("answer: %d\nid_num: %d\nport: %d\n",r,id_num,ntohs(addr.sin_port));
	return(0);
}

